/*
 * gc global variable provides access to GUI Composer infrastructure components and project information.
 * For more information, please see the Working with Javascript guide in the online help.
 */
var gc = gc || {};
gc.services = gc.services || {};


(function() {   // closure for private static methods and data

    var numDevices = 4;
    var partialMessage = [];
    var uartPacketLength = 0;
    var registerInfo;
    var uartDeferred = Q.defer();
    var sequencer = Q();
    
    var tempRegCache;
    var configRegCache;
    var lowRegCache;
    var highRegCache;
    
    var readWriteDropdown;
    var registerPage;
    var registerModel;
    
    var isRead = true;
    
    var setBytes = function(array, size, value, offset)
    {
        for(var i = size; i-- >0; )
        {
            array[offset + i] = value & 0xff;
            value = value >>> 8;
        }
    };
    
    var getResult = function(info, array)
    {
        var numBits = Math.ceil(info.size/8) * 8;
        var mask = (1 << numBits) - 1;
        
        var result = 0;
        for(var i = 0; i < array.length; i++)
        {
            result = (result << 8) | (array[i] & 0xff);
        }
        return result & mask;
    };
    
    Math.convertTemp = function(value) {
         // swap endianness
        value = ((value & 0xFF) << 8) | ((value >> 8) & 0xFF);
        
        if((value >> 15) & 1) {
            // two's complement
            return ((value >> 8) & 0xFF) - 0x100 + 0.0625*((value >> 4) & 0xF);
        }
        return ((value >> 8) & 0xFF) + 0.0625*((value >> 4) & 0xF);
    };
    
    var UARTPacketCodec = function()
    {
    };
    UARTPacketCodec.prototype.rxPktCtr = 0;
    UARTPacketCodec.prototype.rxMsgCtr = 0;
    UARTPacketCodec.prototype.txPktCtr = 0;
    UARTPacketCodec.prototype.txMsgCtr = 0;
    UARTPacketCodec.prototype.strRxBuffer = "";
    UARTPacketCodec.prototype.strTxBuffer = "";
    UARTPacketCodec.prototype.isConnected = false;
    UARTPacketCodec.prototype.initComplete = false;
    UARTPacketCodec.prototype.templateObj = undefined;
    
    UARTPacketCodec.prototype = new gc.databind.IPollingPacketCodec();

    /**
     * Decodes packets from the target device into data objects.  One object for each packet of data.
     * Unless you have chained your custom codec with the CR codec, the packet data is not framed,
     * so this method is responsible for decoding partial packets as well as multiple packets.  Partial packets
     * should not be returned, but deferred until the next decode method call with more raw data to process.  As a
     * result, this method may return 0 or more packets of data on each invocation.
     *
     * @param target -  function to call to pass on the encoded data towards the GUI
     * @param data  - The raw data received that needs to be decoded from packets into objects.
     * @return - true if connected (i.e. valid data received), false if not connected
     */
    UARTPacketCodec.prototype.decode = function(target, data) {
        this.isConnected = true;
        this.rxPktCtr++;
    
        partialMessage = partialMessage.concat(data);
        if(partialMessage.length >= uartPacketLength) {
            
            if(registerInfo !== undefined && readWriteDropdown.selectedText === 'Global' && isRead) {
                
                // cache the response so other device reads can read from the same data
                // don't cache responses for writes
                switch(registerInfo.name.replace(/[0-9]+/, '')) {
                    case 'TEMP_RESULT_':
                        tempRegCache = partialMessage;
                        break;
                    case 'CONFIG_REG_':
                        configRegCache = partialMessage;
                        break;
                    case 'TEMP_LOW_LIMIT_':
                        lowRegCache = partialMessage;
                        break;
                    case 'TEMP_HIGH_LIMIT_':
                        highRegCache = partialMessage;
                        break;
                }
                
                var deviceAddrs = registerInfo.parentGroup.deviceAddrs;
                var packetIndex = 2 + deviceAddrs*2;
            
                if(+registerPage.autoReadValue === -1) {
                    
                    // return device data corresponding to deviceAddrs
                    uartDeferred.resolve(partialMessage.slice(packetIndex, packetIndex + 2));
                    
                    // update register bindings for other three devices
                    for(var i=0; i<numDevices; i++) {
                        packetIndex = 2 + i*2;
                        
                        var bindName = registerInfo.name;
                        bindName = bindName.replace(/[0-9]+/, '');
                        
                        var value = getResult(registerInfo, partialMessage.slice(packetIndex, packetIndex + 2));
                        
                        if(i !== deviceAddrs) {
                            var binding = registerModel.getBinding(bindName + i.toString());
                            binding.updateValue(value);
                        }
                        
                    }
                }
                else {
                    uartDeferred.resolve(partialMessage.slice(packetIndex, packetIndex + 2));
                }
            }
            else {
                uartDeferred.resolve(partialMessage.slice(2,4));
            }
            
            partialMessage = [];
        }
    };
    
    
    /**
     * Initializes target communication and assigns addresses to each device
     *
     * @param settings - UART interface specific settings
     * @return - promise that resolves once response from target is received
     */
    UARTPacketCodec.prototype.connect = function(settings) {
        registerPage = document.querySelector('ti-widget-register-page');
        readWriteDropdown = document.querySelector('#read_write_dropdown');
        registerModel = document.querySelector('ti-model-register').getModel();
        
        registerPage.addEventListener('auto_read_config_changed', function(event) {
            var autoRead = +event.detail.autoRead.value;
            if(autoRead !== -1) {
                // Disable read register buttons if polling
                registerPage.$.read_register.disabled = true;
                registerPage.$.read_all_registers.disabled = true;
            }
            else {
                registerPage.$.read_register.disabled = false;
                registerPage.$.read_all_registers.disabled = false;
            }
        });
        
        var dataBits = settings.dataBits || 16;
        this.dataBitsMask = (1 << dataBits) -1;
        this.dataBitsOffset = settings.dataBitsOffset || 0;
        
        var addrsBits = settings.addrsBits || 2;
        this.addrsBitsMask = (1 << addrsBits) - 1;
        this.addrsBitsOffset = settings.addrsBitsOffset || 1;
        
        // Send initialization command and address assignment
        // 0x55 : Calibration Byte
        // 0x8C : Initialization
        // 0x90 : Set first device address to 0000
        var initData = [];
        setBytes(initData, 3, 0x558C90, 0);
        
        var deferred = Q.defer();
        var that = this;
        sequencer = sequencer.finally(function() {
            uartDeferred = deferred;
    
            uartPacketLength = 3;
            that.encoder(initData);
    
            return deferred.promise;
        });
        return deferred.promise;
        
    };
    
    /**
     * Starts a read operation
     *
     * @param info - identifies which register and which device to read from
     * @return - promise that resolves once response from target is received
     */
    UARTPacketCodec.prototype.readValue = function(info)
    {
        if(readWriteDropdown.selectedText === 'Global') {
            return this.globalRead(info);
        }
        else {
            return this.individualRead(info);
        }
        
    };
    
    /**
     * Starts a write operation
     *
     * @param info - identifies which register and which device to write to
     * @param value - the value to write to the target
     * @return - promise that resolves once response from target is received
     */
    UARTPacketCodec.prototype.writeValue = function(info, value)
    {
        if(readWriteDropdown.selectedText === 'Global') { 
            switch(info.name.replace(/[0-9]+/, '')) {
                case 'TEMP_RESULT_':
                    tempRegCache = undefined;
                    break;
                case 'CONFIG_REG_':
                    configRegCache = undefined;
                    break;
                case 'TEMP_LOW_LIMIT_':
                    lowRegCache = undefined;
                    break;
                case 'TEMP_HIGH_LIMIT_':
                    highRegCache = undefined;
                    break;
            }
            return this.globalWrite(info, value);
        }
        else {
            return this.individualWrite(info, value);
        }
    };
    
    
    UARTPacketCodec.prototype.globalRead = function(info)
    {
        var regAddrs = ((info.addr & this.addrsBitsMask) << this.addrsBitsOffset);
        var command = 0xF1 | regAddrs;
        var globalReadData = [];
        setBytes(globalReadData, 2, 0x5500 | command, 0);
        
        var deferred = Q.defer();
        var that = this;
        sequencer = sequencer.finally(function() {
            var packetCache;
            switch(info.name.replace(/[0-9]+/, '')) {
                case 'TEMP_RESULT_':
                    packetCache = tempRegCache;
                    break;
                case 'CONFIG_REG_':
                    packetCache = configRegCache;
                    break;
                case 'TEMP_LOW_LIMIT_':
                    packetCache = lowRegCache;
                    break;
                case 'TEMP_HIGH_LIMIT_':
                    packetCache = highRegCache;
                    break;
            }
            if(readWriteDropdown.selectedText === 'Global' && +registerPage.autoReadValue !== -1
                && info.parentGroup.deviceAddrs !== (numDevices-1) && packetCache !== undefined) {
                // if polling with global reads, devices 1-n return value from device 0 read cache
                var deviceAddrs = info.parentGroup.deviceAddrs;
                var packetIndex = 2 + deviceAddrs*2;
                
                deferred.resolve(packetCache.slice(packetIndex, packetIndex + 2));
                return deferred.promise;
            }
            
            isRead = true;
            uartDeferred = deferred;
            uartPacketLength = 2 + numDevices*2;
            registerInfo = info;
            that.encoder(globalReadData);
    
            return deferred.promise;
        });
        return deferred.promise;
    };
    
    UARTPacketCodec.prototype.globalWrite = function(info, value)
    {
        var regAddrs = ((info.addr & this.addrsBitsMask) << this.addrsBitsOffset);
        var command = 0xF0 | regAddrs;
        var globalWriteData = [];
        setBytes(globalWriteData, 4, 0x55000000 | (command << 16) | value, 0);
        
        var deferred = Q.defer();
        var that = this;
        sequencer = sequencer.finally(function() {
            isRead = false;
            uartDeferred = deferred;
    
            uartPacketLength = numDevices;
            registerInfo = info;
            that.encoder(globalWriteData);
    
            return deferred.promise;
        });
        return deferred.promise;
    };
    
    UARTPacketCodec.prototype.individualRead = function(info)
    {
        var regAddrs = ((info.addr & this.addrsBitsMask) << this.addrsBitsOffset);
        var deviceAddrs = info.parentGroup.deviceAddrs << 3;
        var command = 0x01 | deviceAddrs | regAddrs;
        var readData = [];
        setBytes(readData, 2, 0x5500 | command, 0);
    
        var deferred = Q.defer();
        var that = this;
        sequencer = sequencer.finally(function() {
            isRead = true;
            uartDeferred = deferred;
            uartPacketLength = 4;
            registerInfo = info;
            
            that.encoder(readData);
            
            return deferred.promise;
        });
        
        return deferred.promise;
    };
    
    UARTPacketCodec.prototype.individualWrite = function(info, value)
    {
        var regAddrs = ((info.addr & this.addrsBitsMask) << this.addrsBitsOffset);
        var deviceAddrs = info.parentGroup.deviceAddrs << 19;
        var command = deviceAddrs | (regAddrs << 16) | value;
        var writeData = [];
        setBytes(writeData, 4, 0x55000000 | command, 0);
    
        var deferred = Q.defer();
        var that = this;
        sequencer = sequencer.finally(function() {
            isRead = false;
            uartDeferred = deferred;
            uartPacketLength = 4;
            registerInfo = info;
            
            that.encoder(writeData);
            
            return deferred.promise;
        });
        
        return deferred.promise;
    };
    
    // Wait for DOMContentLoaded event before trying to access the application template
    var init = function() {
        // You can 'chain' codecs together to make it easier to work with the data
        // coming from the device or going to the device.
        // The following codecs are provided:
        //   CR  - use this if the target data uses /n as a delimiter for framing.
        //         The CR codec will take care of buffering the incoming data so that
        //         your codec will only receive complete 'ready to use' packets
        //  If your target uses /n delimited strings, set baseCodecs = "CR"
        //  If your target uses some other format for data (e.g. binary data or custom strings) leave baseCodecs = "undefined"
        var baseCodecs = "";
        var gcDatabindInitComplete = false;
        var templateInitComplete = false;
        console.log("custom codec: about to register..."); // log a message that you will be able to see in the debugger
        if ((gc) && (gc.databind)) {
            gc.databind.registerCustomCodec("UARTPacketCodec", UARTPacketCodec, baseCodecs);
        } else {
            document.addEventListener('gc-databind-ready',function(){
                gc.databind.registerCustomCodec("custom",UARTPacketCodec, baseCodecs);
                gcDatabindInitComplete = true;
                if (templateInitComplete){
                    UARTPacketCodec.initComplete = true;
                }
            });
        }
    
        UARTPacketCodec.templateObj = document.querySelector('#template_obj');
        if (UARTPacketCodec.templateObj) {
            // Wait for the template to fire a dom-change event to indicate that it has been 'stamped'
            // before trying to access components in the application.
            UARTPacketCodec.templateObj.addEventListener('dom-change',function(){
                if (UARTPacketCodec.initComplete) return;
                UARTPacketCodec.templateObj.async(function(){
                    templateInitComplete = true;
                    if (gcDatabindInitComplete){
                        UARTPacketCodec.initComplete = true;
                    }
                    console.log("Application template has been stamped.");
                    // Now that the template has been stamped, you can use 'automatic node finding' $ syntax to access widgets.
                    // e.g. to access a widget with an id of 'widget_id' you can use CustomCodec.templateObj.$.widgetId
    
                },1);
            });
            
        } else {
           templateInitComplete = true;
           if (gcDatabindInitComplete){
              UARTPacketCodec.initComplete = true;
           }
           console.log("CustomCodec: Failed to find id template_obj in the Application!");
        }
    
    
    };
    
    if (document.querySelector('#template_obj')) {
        init();
    } else {
        document.addEventListener('DOMContentLoaded',init.bind(this));
    }


}());

